// ----------------------------------------------------------------------------
// VFIS 
// ----------------------------------------------------------------------------
// Verified Fredholm Integral Equation Solver
// for linear Fredholm Integral Equations of the second kind
// (Parallel System Version)
// ----------------------------------------------------------------------------
// Verifizierter Integralgleichungsloeser 
// fuer lineare Fredholmsche Integralgleichungen zweiter Art
// (Parallele Systemversion)
// ----------------------------------------------------------------------------
// Copyright (c) 2004-2007 Markus Grimmer
//
// markus.grimmer@math.uni-wuppertal.de
// ----------------------------------------------------------------------------

// This file is part of the Verified Fredholm Integral Equation Solver VFIS.
// 
// VFIS is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// VFIS is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with VFIS; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

/**
 * @file System-FIGL-Loeser-par.hpp
 * VFIS Integral Equation Solver central module header with class definitions.
 * Method/Function definitions see implementation file.
 */

#include <mpi.h>
#include <string>
#include <iostream>    
#include <fstream>    
#include <sstream>
#include <ctime>
#include <vector>

#include "dim2taylor.hpp"
#include "itaylor.hpp"   

#include "vectoroperations.hpp"
#include "taylor_mpicomm.hpp"
#include "vector_mpicomm.hpp"

#include "ilss_par.hpp"

#include "interval.hpp"
#include "ivector.hpp"
#include "intvector.hpp"
#include "imatrix.hpp"
#include "real.hpp"

#include "dot.hpp"
#include "idot.hpp"


using namespace std;
using namespace cxsc;
using namespace taylor;

void start_clock(clock_t&);
void print_time_used(clock_t, ofstream&);

interval Imid(interval&);
itaylor blow(itaylor&, real);
vector<itaylor> blow(vector<itaylor>&, real);
ivector Powers (interval, int);
void Systembasis_Integration(imatrix&, imatrix&,
                             int, int, ivector);

/**
 * @class IntegralSystemOperatorRG
 * Integral System Operator specialized for the Taylor expansion remainder terms 
 * ("Restglied" in German, that's where "RG" comes from...)
 *
 * Application of an integral (system) operator K to a function vector y: 
 *
 * (Ky)(s) = Int_a^b k(s,t)y(t)dt
 */
class IntegralSystemOperatorRG
{
  private:
    /** 
     * Kernel coefficients from remainder denotation of the kernel function.
     */
    vector<vector<ivector> > KSys0;
    /** 
     * Domain of the (free) variable
     */
    ivector      SSys;
    /** 
     * Domain midpoint
     */
    ivector      SmidSys;

  public:
    IntegralSystemOperatorRG(const IntegralSystemOperatorRG&);
    IntegralSystemOperatorRG(vector<vector<ivector> >&, ivector&, ivector&);
    vector<itaylor> operator()(vector<itaylor>&, int, int,
                               imatrix&, imatrix&); 
};

// -----------------------------------------------------------------------------

vector<itaylor> IterativesVerfahren (IntegralSystemOperatorRG&, 
                             vector<itaylor>&, 
                             int, int, 
                             imatrix&, imatrix&, int&, ofstream&);

// -----------------------------------------------------------------------------

/**
 * @class IntegralGleichungsSystem 
 * The integral equation system including all 
 * directly associated data and functions as attributes and methods.
 *
 * Not included: Indirectly associated functions (iteration functions, etc.)
 */
class IntegralGleichungsSystem
{
  private:
    /**
     * Typedef for kernel function pointer.
     */
    typedef dim2taylor (*KFkt)(dim2taylor_vector&);
    /**
     * Typedef for right hand side function pointer.
     */
    typedef itaylor (*ReFkt)(itaylor&);
    /**
     * Kernel function pointer matrix.
     */
    vector<vector<KFkt> > KernFktSys0;
    /**
     * Right hand side function pointer vector.
     */
    vector<ReFkt> RechtsFktSys0;

    /**
     * Kernel Taylor coefficients
     */
    vector<vector<dim2taylor> > KernSys0;
    /**
     * Right hand side Taylor coefficients
     */
    vector<itaylor> RSSys0;
        
    /**
     * @name Domains and domain midpoints
     */
    //@{
    interval     Sglob, Tglob, RS_Sglob;
    ivector      SSys, TSys, SmidSys, TmidSys;
    ivector      RS_SSys, RS_SmidSys;
    //@}
    /**
     * Linear system solution.
     */
    ivector         ILGS_Lsg;
    /**
     * Integral equation system solution.
     */
    vector<itaylor> Lsg0;
    /**
     * Integral equation system order.
     */
    int sysord;
    /**
     * Taylor order.
     */
    int ord;
    /**
     * Integrals of monomials.
     */
    imatrix BasisInt_a;
    /**
     * Integrals of monomials.
     */
    imatrix BasisInt_b;
    /**
     * Remainders of kernel Taylor expansions
     */
    vector<vector<ivector> > RestkernSys0;

  public:

    /**
     * @name Getters
     * Getters (Read access functions).
     */
    //@{
    vector<vector<dim2taylor> >& getKern() { return KernSys0; }
    dim2taylor& getKern(int i, int j) { return (KernSys0.at(i)).at(j); }
    vector<itaylor>& getRS()          { return RSSys0; }
    itaylor& getRS(int i)             { return RSSys0.at(i); }
    ivector& getSSys()      { return SSys; }    
    ivector& getTSys()      { return SSys; }    
    ivector& getRS_SSys()   { return RS_SSys; }
    interval getS()         { return Sglob; }
    interval getS(int i)    { return SSys[i+1]; }     // Indexbereich "0..n-1"
    interval getSmid(int i) { return SmidSys[i+1]; }  // Indexbereich "0..n-1"
    interval getT()         { return Tglob; }
    interval getT(int i)    { return TSys[i+1]; }     // Indexbereich "0..n-1"
    interval getTmid(int i) { return TmidSys[i+1]; }  // Indexbereich "0..n-1"
    interval getRS_S()        { return RS_Sglob; }
    vector<itaylor>& getLsg() { return Lsg0; }
    itaylor& getLsg(int i)    { return Lsg0.at(i); }
    //@}
  
    IntegralGleichungsSystem (int sysord=0, int ord=0);
    IntegralGleichungsSystem (KFkt,interval, interval,
                              ReFkt, interval, int, int, ofstream&, string);  

    void PartInitIntegralGleichungsSystem (KFkt,interval, interval,
                              ReFkt, interval, int, int, ofstream&, string); 
    int DistributeValues(int);
 
    void Einschliessung_berechnen(ofstream&,string,int,int,int&);
    void SystemLoesungsansatz(vector<vector<vector<itaylor> > >&,
                              vector<itaylor>&, int, ofstream&, string, 
                              int, int, int, int&);
    void SystemLoesung_bilden(vector<vector<vector<itaylor> > >&,
                              vector<itaylor>&, int, ofstream&);
    real Genauigkeit(int);

    void Export(string dest, string fname, string varname="IGLSys_Lsg");
    void Export_Plot(string dest, string fname, int steps, 
                     string varname="IGLSys_Lsg", 
                     string plottitle="Integral Equation System Solution Enclosure");

    
    friend ostream& operator<< (ostream&, IntegralGleichungsSystem&);

};


