// ----------------------------------------------------------------------------
// ILSS_PAR 
// ----------------------------------------------------------------------------
// Parallel Interval Linear System Solver
// ----------------------------------------------------------------------------
// Paralleler Verifizierter Loeser fuer lineare Intervallgleichungssysteme
// ----------------------------------------------------------------------------
// This Version: 
// Copyright (C) 2004-2007 Markus Grimmer
// Based on: 
// Serial Version Copyright (C) 2003 Carlos Holbig and Walter Kraemer
//
// markus.grimmer@math.uni-wuppertal.de
// ----------------------------------------------------------------------------

// This file is part of the Parallel Interval Linear System Solver ILSS_PAR.
// 
// ILSS_PAR is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// ILSS_PAR is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with ILSS_PAR; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

/**
 * @file ilss_par.cpp
 * 
 * ILSS_PAR Parallel Interval Linear System Solver 
 * Central module with all central methods/functions.
 * 
 * Parallel Version Copyright (C) 2004-2007 Markus Grimmer
 *
 * Based on: 
 * Serial Version Copyright (C) 2003 Carlos Holbig and Walter Kraemer
 *
 * Parallel verified solution of interval linear systems
 * using Rump's algorithm.
 */

#include <mpi.h>

#include <iostream> 
#include <fstream> 
#include <vector>
#include <iomanip>
#include <imatrix.hpp>
#include <mvi_util.hpp>
#include "ilss_par.hpp"

using namespace cxsc;
using namespace std;

/**
 * @name Constants
 */
//@{
const real zerotest = 1e6;
const real delta    = 1e-15;
const real eps1     = 1e-15;
const real sqrt_01  = 0.31622777;
const real limit    = 1e20; 
//@}

static int m, n, dim;

// ----------------------------------------------------------------------------

/**
 * Zero vector
 */
static ivector null(ivector A)
{
  return (A = 0.0);
}

/**
 * Transposition of a matrix
 *
 * @param A The matrix
 * @return The transposed matrix
 */
static imatrix transp ( imatrix& A )
{                                            
  int      n;
  imatrix  res(Lb(A,2),Ub(A,2),Lb(A,1),Ub(A,1));

  for (n = Lb(A,1); n <= Ub(A,1); n++) Col(res,n) = Row(A,n);
  return res;
}

/**
 * REL computes componentwise the maximum relative error of A w.r.t B. 
 * if A[i] and B[i] do not have the same sign or if B[i] = 0, then 
 * rel. error = 0 for this component.
 * @param A The new value of an iteration 
 * @param B The old value of an iteration
 * @return Relative error
 */
real REL(rvector A, rvector B)
{
  int i;
  real p,r,ai,bi;
 
  p=0;
  for(i=Lb(A);i<=Ub(A);i++) // A,B must have same index range
  {
    ai = A[i];
    bi = B[i];
    if( ai*bi <= 0.0 || zerotest*abs(ai) < abs(bi) ) r = 0.0;
      else r = abs( (ai-bi)/bi );
    if (r>p) p = r;
  }
  return p;
} // end REL

/**
 * TOO_BAD = accuracy of A is far too bad
 * 
 * note: 0 for false, 1 for true;
 *
 * @param A A vector
 * @return Too bad or not
 */
bool TOO_BAD(ivector &A)
{
  int  i;
  bool bad;
 
  bad = false;
  for(i=Lb(A);i<=Ub(A);i++)
  {
    bad = bad || Inf(A[i]) < -limit && Sup(A[i]) > limit;
  }
  return bad;
} // end TOO_BAD

/**
 * x1 is the new, x0 the old value of an iteration. If a component of x1
 * has decreased by more than a factor of zerotest, then this component
 * is set to 0. The same is done if the sign of a component has changed.
 * 
 * @param  x0 old value of iteration
 * @param  x1 new value of iteration
 */
void GUESS_ZEROES(rvector& x0, rvector& x1)
{
  int  i;
  real MAXX;
 
  MAXX = 0.0;
  for(i=Lb(x1);i<=Ub(x1);i++)
    if(abs(x1[i])>MAXX)
      for(i=Lb(x1);i<=Ub(x1);i++)
        if( x0[i]*x1[i] < 0.0 || zerotest*abs(x1[i]) < abs(x0[i]) 
	        && MAXX > zerotest*abs(x1[i]) )
          x1[i] = 0.0;
} // end GUESS_ZEROES

// ----------------------------------------------------------------------------

/**
 * Parallel linear system solver (square matrix expected).
 *
 * The result of Y is an enclosure of the solution
 * errc = 0: Y is enclosure of the solution
 * errc = 1: no enclosure obtained, bad condition (?)
 * errc = 2: no enclosure obtained, matrix A singular (?)
 *
 * FLAGS = 0: R1,R2,C have not yet been computed
 * FLAGS = 1: only R1 has been computed
 * FLAGS = 2: R1 and corresponding C have been computed
 * FLAGS = 3: R1 and R2 are computedbut not the corresponding C
 * FLAGS = 4: R1, R2 and the corresponding C have been computed
 *
 * This version: 
 * Includes parameters for already distributed data 
 * (matrix columns) and already initialized parallel environment
 *
 * @param[in]  A  The matrix of the system
 * @param[in]  b  Right hand side
 * @param[out] Y  Solution enclosure
 * @param[in]  MyA       Distributed A
 * @param[in]  mycolumns Local columns
 * @param[in]  mycol     Local to global column mapping
 * @param[in]  distribute Distribution of columns necessary?
 * @param[in]  procs      Number of parallel processes
 * @param[in]  mypid      Process ID
 * @param[out] errc       Error code
 * @param[out] commerrc   Error code
 * @param[in]  R1         Real matrix
 * @param[in]  R2         Real matrix
 * @param[in]  C          Interval Matrix
 * @param[out] FLAGS      Flag
 * @param[in,out] ausg    Output file stream
 */
void LSS(imatrix& A, ivector& b, ivector& Y,  
         imatrix& MyA, vector<int>& mycolumns, intvector& mycol, 
         bool distribute, int procs, int mypid, int& errc, int& commerrc,
         rmatrix& R1, rmatrix& R2, imatrix& C, int& FLAGS, ofstream& ausg)
{

  rmatrix       D( Lb(A,1),Ub(A,1),Lb(A,2),Ub(A,2) ), 
  	            R1temp( Lb(A,1),Ub(A,1),Lb(A,2),Ub(A,2) ),
                    R2temp( Lb(A,1),Ub(A,1),Lb(A,2),Ub(A,2) );
  rvector       x0( Lb(A,1),Ub(A,1) ), x1( Lb(A,1),Ub(A,1) ), 
	            y0( Lb(A,1),Ub(A,1) ), x1temp( Lb(A,1),Ub(A,1) ); 
  ivector       Y1( Lb(A,1),Ub(A,1) ), YA( Lb(A,1),Ub(A,1) ), 
	            Z( Lb(A,1),Ub(A,1) ),Y1temp( Lb(A,1),Ub(A,1) );
  int           i,j,k,err;
  bool          ready;
  int           readyint;
  real          p,bound,eps1;
  dotprecision  accu;
  idotprecision iaccu;
  interval      itemp;
  rmatrix       AM( Lb(A,1),Ub(A,1),Lb(A,2),Ub(A,2) );
  rvector       bm( Lb(b),Ub(b) );

  rmatrix       MyAM( Lb(MyA,1),Ub(MyA,1),Lb(MyA,2),Ub(MyA,2) );

  rmatrix       RAtmp; 
  rmatrix       R1AM;

  int m = Ub(A,ROW) - Lb(A,ROW) + 1;
  int n = Ub(A,COL) - Lb(A,COL) + 1;
  if (m!=n) 
  {
    err=5;
    return;
  }

  ausg << "Starting LSS..." << endl;

  //----------------------------------------------------------------------------
  // MPI preparation
  //----------------------------------------------------------------------------

  MPI_Status status;

  // mypid/procs already initialized
  
  int root=0, rootmul=1; 

  // Remark 1: Up to now, root=0 is expected. Some modifications might 
  //           have to be done if you need root!=0.
  // Remark 2: matmul expects to operate on a contiguous set of proc id's

  //----------------------------------------------------------------------------
  
  if (mypid==root)
  {  
    AM=mid(A); bm=mid(b);
  }
  
  MyAM=mid(MyA);

  ready = false;
  
  if (Lb(A,1)!=Lb(A,2))
  {
    err=5;
    return;
  }
  
  // begin USE_SINGLE_R - compute approximate inverse of A

  commerrc=MPI_Bcast(&FLAGS,1,MPI_INT,root,MPI_COMM_WORLD);

  if (FLAGS<3)
  {
    err = 0;
    if (FLAGS<1)
    {
      RAtmp = AM;
                   
      //----------------------------------------------------------------------------
      // Parallel computation of approximate inverse
      //----------------------------------------------------------------------------

      if (distribute)
      {
        // Inversion with column distribution
        MatInv( RAtmp, R1, procs, mypid, err, commerrc, ausg );
      }
      else
      {
        // Inversion without column distribution
              MatInv( RAtmp, R1, MyAM, mycolumns, mycol, 
                false, procs, mypid, err, commerrc, ausg );
      }
      
      // -> Now R1 on proc. #root is the approximate inverse.
      // Procs #root+1..n don't have the full R1!
      
      if (err==0) FLAGS = 1;
    }

    commerrc=MPI_Bcast(&err,1,MPI_INT,0,MPI_COMM_WORLD);
    commerrc=MPI_Bcast(&FLAGS,1,MPI_INT,0,MPI_COMM_WORLD);

    if (err==0)
    {

       if (mypid==root)
       {
         // Send R1 and A to proc. #rootmul

         if (FLAGS<2)
         {
           commerrc=MPI_Send(R1, rootmul, 0, MPI_COMM_WORLD);
           commerrc=MPI_Send(A, rootmul, 0, MPI_COMM_WORLD);
         }

         //------------------------------------------------------
         // Proc. #0:
         // floating point defect iteration: result is x1
         //------------------------------------------------------

         bound = 100.0*sqrt_01;
         x1 = R1*bm;
         k = 0;
         do
         // iterate x = x + R*(b-Ax)
         {
           k = k + 1;
           x0 = x1;
 
           // x1 := #*(bm - AM*x0)
           for (i=Lb(x0);i<=Ub(x0);i++)
           {
             accu = bm[i];
             accumulate(accu,-AM[Row(i)],x0);
             x1[i] = rnd(accu);
           }
 
           // x1 := #*(x0 + R1*x1)
           for (i=Lb(x1);i<=Ub(x1);i++)
           {
             accu = x0[i];
             accumulate(accu,R1[Row(i)],x1);
             x1temp[i] = rnd(accu);
           }
 
           x1 = x1temp;
           p = REL(x1,x0);
           if(k>5) bound = bound*sqrt_01;
         } while ((p<bound || k<=5) && p>=delta);
 
         GUESS_ZEROES(x1,x0);
 
         //------------------------------------------------------
         // Proc. #root:
         // Compute enclosure y0+y1 of the residuum b-A*x1 
         // of the aproximation x1
         // and initialize Y1:=Z:= R1*(b-A*x1).
         //
         // In the meantime C:= I-R1*A is computed by the other 
         // procs.
         //------------------------------------------------------

         // y0 := #*(b-A*x1) = modified to y0:=MID(##(b - A*x1))
         for (i=Lb(x1);i<=Ub(x1);i++)
         {
           iaccu = b[i];
           accumulate(iaccu,-A[Row(i)],x1);
           rnd(iaccu,itemp);
           y0[i] = mid(itemp);
         }
 
         // Y1 := ##(b-A*x1-y0)
         for (i=Lb(x1);i<=Ub(x1);i++)
         {
           iaccu = b[i];
           accumulate(iaccu,-A[Row(i)],x1);
           iaccu = iaccu - y0[i];
           rnd(iaccu,Y1[i]);
         }
 
         // Y1 := ##( R1*y0 + R1*Y1 );
         for (i=Lb(R1,1);i<=Ub(R1,1);i++)
         {
           iaccu = 0.0;
           accumulate(iaccu,R1[Row(i)],y0);
           accumulate(iaccu,R1[Row(i)],Y1);
           rnd(iaccu,Y1temp[i]);
         }
 
         Y1 = Y1temp;
         Z = Y1;
       }
       //---------------
       //end mypid==root  
       //---------------
       else
       {
         //------------------------------------------------------
         // Proc. #rootmul..n:
         // Compute C:= I-R1*A.
         //------------------------------------------------------
         if (FLAGS<2)
         {
           if (mypid==rootmul)
           {
             // Receive R1 and A from proc. #root
             
             commerrc=MPI_Recv(R1, root, 0, MPI_COMM_WORLD, &status);
             commerrc=MPI_Recv(A, root, 0, MPI_COMM_WORLD, &status);

             IminusAB(R1,A,C,n,n,n,rootmul,rootmul,procs-1,err,commerrc,ausg);

             // Send C back to root
             commerrc=MPI_Send(C, root, 0, MPI_COMM_WORLD);
           }
           else if (mypid>rootmul)
           {
             IminusAB(R1,A,C,n,n,n,rootmul,rootmul,procs-1,err,commerrc,ausg);
           }
           FLAGS = 2;
         }// end if FLAGS<2
       }
       //---------------------
       //end else (mypid!=root)
       //---------------------

       if (mypid==root)
       {
         // Receive C from rootmul
         commerrc=MPI_Recv(C, rootmul, 0, MPI_COMM_WORLD, &status);

         commerrc=MPI_Send(AM, rootmul, 0, MPI_COMM_WORLD);
         FLAGS = 2;
       
         if (Z==null(Z))
         {
           Y = x1; // exact solution! (however, not necessarily unique!)
           errc = 0;
           ready = true;
         }
         else
         {
           // interval iteration until inclusion is obtained (or max. iteration count)
           k = 0;
           eps1 = 0.1;
           do
           {
             if (k>=5) eps1 = 5*eps1;
             k = k+1;
             YA = Blow(Y1,eps1);
             Y1 = Z + C*YA;
             ready = in(Y1,YA);
           } while ( !ready && k<10 && !TOO_BAD(Y1) );

           // result
           if (ready)
           {
             Y = x1 + Y1;
             errc = 0;
           }
         }
       }
       //---------------
       //end mypid==root
       //---------------
       else
       {
         
         //----------------------------------------------------------------------
         // Parallel: R1AM=R1*AM
         //----------------------------------------------------------------------
         
         if (mypid==rootmul)
         {
           // R1 is already available on proc. #rootmul

           commerrc=MPI_Recv(AM, root, 0, MPI_COMM_WORLD, &status);

           MatMul(R1,AM,R1AM,n,n,n,rootmul,rootmul,procs-1,err,commerrc,ausg);

         }
         else if (mypid>rootmul)
         {
           MatMul(R1,AM,R1AM,n,n,n,rootmul,rootmul,procs-1,err,commerrc,ausg);
         }
       }
       //----------------------
       //end else (mypid!=root)
       //----------------------
    }// if err==0
    else ready = false;
    
    
  } // end USE_SINGLE_R
 
//------------------------------------------------------------------------------
// Part II
//------------------------------------------------------------------------------
// if no success: try again with approximate inverse R = R1+R2 of double length
// USE_DOUBLE_R to try again with approximate inverse R = R1+R2 of double length 
//------------------------------------------------------------------------------

 readyint=(ready==true)?1:0;
 commerrc=MPI_Bcast(&readyint,1,MPI_INT,root,MPI_COMM_WORLD);
 if (mypid!=root)
 {
   ready=(readyint==0)?false:true;
 }

 if  (!ready)
  {

    ausg << "LSS: Proc. " << mypid << ": Step 1 not successful. Starting step 2..." << endl;

    err = 0;
    if (FLAGS<3)
    {
      // R1AM = R1*AM; has been computed previously and is available 
      // on proc. #rootmul

      if (mypid==root)
      {
        commerrc=MPI_Recv(R1AM, rootmul, 0, MPI_COMM_WORLD, &status);
      }
      else if (mypid==rootmul)
      {
        commerrc=MPI_Send(R1AM, root, 0, MPI_COMM_WORLD);
      }
      
      //----------------------------------------------------------------------------
      // Parallel computation of approximate inverse
      //----------------------------------------------------------------------------

      MatInv( R1AM, R2, err, commerrc, ausg );
      
      // -> Now R2 on proc. #root is the approximate inverse.
      // Procs #root+1..n don't have the full R2!

      commerrc=MPI_Bcast(&err,1,MPI_INT,0,MPI_COMM_WORLD);
      commerrc=MPI_Bcast(&FLAGS,1,MPI_INT,0,MPI_COMM_WORLD);
      
      if (err==0)
      {

        FLAGS = 3;

        //----------------------------------------------------------------------
        // Parallel: "D := #*(R2*R1)"
        // Parallel: "R2 := #* (R2*R1 - D);"
        //----------------------------------------------------------------------

        if (mypid==root)
        {
          // Proc. #root has R1 and R2!

          ABminusRndAB(R2,R1,R1temp,R2temp,n,n,n,
                       root,root,procs-1,err,commerrc,ausg);

          R1 = R1temp;
          R2 = R2temp;

          // -> Now R1 and R2 on proc. #root 
          // are the desired result "R1+R2".
          // Procs #root+1..procs-1 don't have the full results!
 
        }
        else if (mypid>root)
        {
          ABminusRndAB(R2,R1,R1temp,R2temp,n,n,n,
                       root,root,procs-1,err,commerrc,ausg);
        }
        
      }
    }

    commerrc=MPI_Bcast(&err,1,MPI_INT,0,MPI_COMM_WORLD);
    commerrc=MPI_Bcast(&FLAGS,1,MPI_INT,0,MPI_COMM_WORLD);

    if (err==0)
    {
      if (mypid==root)
      {
           commerrc=MPI_Send(R1, rootmul, 0, MPI_COMM_WORLD);
           commerrc=MPI_Send(R2, rootmul, 0, MPI_COMM_WORLD);

         //------------------------------------------------------
         // Proc. #0:
         // floating point defect iteration: result is x1+x0
         //------------------------------------------------------

         bound = 100.0*sqrt_01;
 
         // x1 := #*( R1*bm + R2*bm );
         for (i=Lb(R1,1);i<=Ub(R1,1);i++)
         {
           accu = 0.0;
           accumulate(accu,R1[Row(i)],bm);
           accumulate(accu,R2[Row(i)],bm);
           x1[i] = rnd(accu);
         }
 
         // x0 = #*( R1*bm + R2*bm - x1):
         for (i=Lb(R1,1);i<=Ub(R1,1);i++)
         {
           accu = -x1[i];
           accumulate(accu,R1[Row(i)],bm);
           accumulate(accu,R2[Row(i)],bm);
           x0[i] = rnd(accu);
         }
 
         k = 0;
         do
         // iteration x = x + (R1+R2)*(b-Ax), x = x1 + x0)
         {
           k = k+1;
 
           // y0 = #*(bm - AM*x1 - AM*x0)
           for (i=Lb(A,1);i<=Ub(A,1);i++)
           {
             accu = bm[i];
             accumulate(accu,-AM[Row(i)],x1);
             accumulate(accu,-AM[Row(i)],x0);
             y0[i] = rnd(accu);
           }
 
           // y0 := #*(x0 + R1*y0 + R2*y0);
           for (i=Lb(R1,1);i<=Ub(R1,1);i++)
           {
             accu = x0[i];
             accumulate(accu,R1[Row(i)],y0);
             accumulate(accu,R2[Row(i)],y0);
             x1temp[i] = rnd(accu);
           }
           y0 = x1temp;
 
           p = REL (x1+y0,x1+x0);
           y0 = x1 + y0;
 
           // x0 := #*(x1 + x0 - y0)
           for (i=Lb(x1);i<=Lb(x1);i++)
           {
             accu = x1[i] + x0[i] - y0[i];
             x0[i] = rnd(accu);
           }
               x1 = y0;
 
                   if (k>5) bound = bound * sqrt_01;
         } while ( (p<bound || k<=5) && p>=delta );

         // compute enclosure y0+Y1 of the residuum b-A*x1 of the approximation x1 and
         // initialize Y1:= Z:= (R1+R2)*(b-A*x1), C:= I-(R1+R2)*A
 
         // y0 := #*(b-A*x1)  ====> y0:=MID(##(b - A*x1))
         for (i=Lb(x1);i<=Ub(x1);i++)
         {
           iaccu = b[i];
           accumulate(iaccu,-A[Row(i)],x1);
           rnd(iaccu, itemp);
           y0[i] = mid(itemp);
         }
 
         // Y1 := ##(b-A*x1-y0)
         for (i=Lb(x1);i<=Ub(x1);i++)
         {
           iaccu = b[i];
           accumulate(iaccu,-A[Row(i)],x1);
           iaccu = iaccu - y0[i];
           rnd(iaccu,Y1[i]);
         }
 
         // Y1 := ##(R1*y0 + R2*y0 + R1*Y1 + R2*Y1 );
         for (i=Lb(R1,1);i<=Ub(R1,1);i++)
         {
           accu = 0.0;
           accumulate(accu,R1[Row(i)],y0);
           accumulate(accu,R2[Row(i)],y0);
           iaccu = accu;
           accumulate(iaccu,R1[Row(i)],Y1);
           accumulate(iaccu,R2[Row(i)],Y1);
           rnd(iaccu,Y1temp[i]);
         }
         Y1 = Y1temp;
         Z = Y1;
      }
      //---------------
      //end mypid==root 
      //---------------
      else
      {
         //------------------------------------------------------
         // Proc. #rootmul..n:
         // Compute C:= ## (ID(A) - R1*A - R2*A);
         //------------------------------------------------------

         if (FLAGS<4)
         {
           if (mypid==rootmul)
           {
             // Receive R1 and R2 from proc. #root

             commerrc=MPI_Recv(R1, root, 0, MPI_COMM_WORLD, &status);
             commerrc=MPI_Recv(R2, root, 0, MPI_COMM_WORLD, &status);

             IminusA1A2B(R1,R2,A,C,n,n,n,rootmul,rootmul,procs-1,err,commerrc,ausg);

             commerrc=MPI_Send(C, root, 0, MPI_COMM_WORLD);
           }
           //------------------
           //end mypid==rootmul
           //------------------
           else if (mypid>rootmul)
           {
             IminusA1A2B(R1,R2,A,C,n,n,n,rootmul,rootmul,procs-1,err,commerrc,ausg);
           }
           FLAGS = 4;
         }// end if FLAGS<4
      }
      
      if (mypid==root)
      {
         // Receive C from rootmul
         commerrc=MPI_Recv(C, rootmul, 0, MPI_COMM_WORLD, &status);

         if (Z==null(Z))
         {
           Y = x1; // exact solution! (however, not necessarily unique!)
           errc = 0;
           ready = true;
         }
         else
         {
           // interval iteration until inclusion is obtained (or max. iteration count)
           k = 0;
           eps1 = 0.1;
           do
           {
             if (k>=5) eps1 = 5*eps1;
             k = k+1;
             YA = Blow( Y1, eps1);
             Y1 = Z + C*YA;
             ready = in(Y1,YA);
           } while ( !ready && k<10 && !TOO_BAD(Y1) );

           // output of the result
           if (ready)
           {
             Y = x1 + Y1;
             errc = 0;
           }
           else errc = 1;
         }
           
      }
      //---------------
      //end mypid==root 
      //---------------
           
    }//end if err==0
    else errc = 2;
  } // end USE_DOUBLE_R

  ausg << "LSS finished." << endl;
  
  
}  // end LSS

// ----------------------------------------------------------------------------

/**
 * Parallel linear system solver (square matrix expected).
 *
 * This version: 
 * Wrapper without local variables.
 * Includes parameters for already distributed data 
 * (matrix columns) and already initialized parallel environment
 * 
 * @param[in]  A  The matrix of the system
 * @param[in]  b  Right hand side
 * @param[out] Y  Solution enclosure
 * @param[in]  MyA Distributed A
 * @param[in]  mycolumns Local column subset
 * @param[in]  mycol     Local to global column mapping
 * @param[in]  distribute Distribution of columns necessary?
 * @param[in]  procs      Number of parallel processes
 * @param[in]  mypid      Process ID
 * @param[out] errc       Error code
 * @param[out] commerrc   Error code
 * @param[in,out] ausg    Output file stream
 */
void LSS( imatrix& A, ivector& b, ivector& Y, 
          imatrix& MyA, vector<int>& mycolumns, intvector& mycol, 
          bool distribute, int procs, int mypid, int& errc, int& commerrc,
          ofstream& ausg) 
{
  rmatrix       R1(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW)),
                R2(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW));
  imatrix       C(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW));
  idotprecision Accu;
  int           FLAGS;

  if( dim == 1 )     // Treat trivial case separately
    if( in(0.0,A[Lb(A,ROW)][Lb(A,COL)]) ) errc = 2;
      else
	  {
        Y[Lb(Y)] = b[Lb(b)] / A[Lb(A,ROW)][Lb(A,COL)];
        errc = 0;
      }
    else
    {
      FLAGS = 0;
      LSS( A, b, Y, MyA, mycolumns, mycol, distribute, procs, mypid,
           errc, commerrc, R1, R2, C, FLAGS, ausg );
    }
} // end

// ----------------------------------------------------------------------------

/**
 * Parallel linear system solver (square matrix expected).
 * 
 * This version: 
 * Wrapper without local variables.
 * Without parameters for distributed data 
 * -> Distribution will be done inside the routine.
 * 
 * Includes parameters for already initialized parallel environment.
 *
 * @param[in]  A  The matrix of the system
 * @param[in]  b  Right hand side
 * @param[out] Y  Solution enclosure
 * @param[in]  procs      Number of parallel processes
 * @param[in]  mypid      Process ID
 * @param[out] errc       Error code
 * @param[out] commerrc   Error code
 * @param[in,out] ausg    Output file stream
 */
void LSS(imatrix& A, ivector& b, ivector& Y,  
         int procs, int mypid, int& errc, int& commerrc, ofstream& ausg)
{
  vector<int> mycolumns;
  intvector mycol;
  imatrix MyA;
   
  LSS(A,b,Y,MyA,mycolumns,mycol,true,procs,mypid,errc,commerrc,ausg);
}

// ----------------------------------------------------------------------------

/**
 * Parallel linear system solver (square matrix expected).
 *
 * This version: 
 * Wrapper without local variables.
 * Without parameters for distributed data and parallel environment.
 * -> Distribution will be done inside the routine.
 * -> Parallel environment will be initialized 
 *    inside the routine.
 *
 * @param[in]  A  The matrix of the system
 * @param[in]  b  Right hand side
 * @param[out] Y  Solution enclosure
 * @param[out] errc       Error code
 * @param[out] commerrc   Error code
 * @param[in,out] ausg    Output file stream
 */
void LSS(imatrix& A, ivector& b, ivector& Y,  
         int& errc, int& commerrc, ofstream& ausg)
{
  int procs;
  int mypid;

  commerrc=MPI_Comm_rank(MPI_COMM_WORLD, &mypid);
  commerrc=MPI_Comm_size(MPI_COMM_WORLD, &procs);

  LSS(A,b,Y,procs,mypid,errc,commerrc,ausg);

}

// ----------------------------------------------------------------------------
