// ----------------------------------------------------------------------------
// ILSS_PAR 
// ----------------------------------------------------------------------------
// Parallel Interval Linear System Solver
// Matrix multiplication/operation module
// ----------------------------------------------------------------------------
// Paralleler Verifizierter Loeser fuer lineare Intervallgleichungssysteme
// Modul fuer Matrixoperationen
// ----------------------------------------------------------------------------
// Copyright (C) 2004-2007 Markus Grimmer
//
// markus.grimmer@math.uni-wuppertal.de
// ----------------------------------------------------------------------------

// This file is part of the Parallel Interval Linear System Solver ILSS_PAR.
// 
// ILSS_PAR is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// ILSS_PAR is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with ILSS_PAR; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

/**
 * @file matmul_par.cpp
 * 
 * ILSS_PAR Parallel Interval Linear System Solver 
 * Matrix multiplication/operation module.
 * 
 * Copyright (C) 2004-2007 Markus Grimmer
 *
 * Contains a number of parallel matrix operations 
 * (including matrix multiplication) for 
 * the parallel verified linear system solver ILSS_PAR.
 * 
 */


#include <mpi.h>

#include <rmatrix.hpp>  
#include <imatrix.hpp>  
#include <dot.hpp>      
#include <idot.hpp>     
#include <intvector.hpp>
#include <intmatrix.hpp>
#include <mvi_util.hpp> 

#include <cmath>

#include <vector>
#include <iterator>
#include <algorithm>
#include <iostream>     

#include "cxsc_mpicomm_templ.hpp" // MPI Communication for C-XSC
#include "vectoroperations.hpp"   // Additional ops for STL vectors
#include "matmul_par.hpp"         // Local header


using namespace cxsc;
using namespace std;

// ----------------------------------------------------------------------------

/**
 * Matrix multiplication imatrix*imatrix A:r x s, B:s x t
 * 
 * @param[in]   A         Input matrix
 * @param[in]   B         Input matrix
 * @param[out]  R         Result
 * @param[in]   r         Matrix dimension 1
 * @param[in]   s         Matrix dimension 2
 * @param[in]   t         Matrix dimension 3
 * @param[in]   root      Master process
 * @param[in]   startproc ID of first involved proc (to date: expected to be =root)
 * @param[in]   endproc   ID of last involved proc
 * @param[out]  errc      Error code
 * @param[out]  commerrc  Error code
 * @param[in,out] ausg    Output file stream
 */
void MatMul(imatrix& A, imatrix& B, imatrix& R, int r, int s, int t,
            int root, int startproc, int endproc,  
            int& errc, int& commerrc, ofstream& ausg) 
{
  //----------------------------------------------------------------------------
  // MPI preparation
  //----------------------------------------------------------------------------

  MPI_Status status;
  int procs=endproc-startproc;
  int mypid;

  commerrc=MPI_Comm_rank(MPI_COMM_WORLD, &mypid);

  ausg << "Start Matrix multiplication..." << endl;

  const real eps = 1e-15;
  bool ok=true;

  SetLb(A,1,1);
  SetLb(A,2,1);
  SetLb(B,1,1);
  SetLb(B,2,1);

  if (mypid==root)
    Resize(R,1,r,1,t); // Result matrix        
  
  imatrix MyA, MyB, MyR;

  if (mypid==root)
    ok = ((s==(Ub(A,2)-Lb(A,2)+1))&&(s==(Ub(B,1)-Lb(B,1)+1)))? true : false; 
  
  if ((root<startproc)||(root>endproc))
  { 
    commerrc=MPI_ERR_DIMS; 
    return;
  }

  //----------------------------------------------------------------------------
  // Column distribution:
  //----------------------------------------------------------------------------

  // Determine number of procs for rectangular block subdivision of matrices.

  int p1=(int)floor(sqrt((double)procs));
  int p2=p1;
  while (p1*(p2+1)<=procs) 
    p2++;
  endproc=startproc+p1*p2-1;
  
  if ((mypid<startproc)||(mypid>endproc))
    return;
  
  int myproccol=(mypid-startproc)/p1;
  int myprocrow=((mypid-startproc)%p1);

  intvector startrows(0,p1-1);
  intvector endrows(0,p1-1);
  intvector startcols(0,p2-1);
  intvector endcols(0,p2-1);
  
  int mystartrow, myendrow, mystartcol, myendcol;

  // Determine startrows, endrows, startcolumns, endcolumns

  for (int i=0; i<p1; i++)
  {
    if (i<=r%p1) 
      startrows[i]=1+i*((r/p1)+1);
    else
      startrows[i]=1+(r%p1)*((r/p1)+1)+(i-(r%p1))*(r/p1);
    if (i<r%p1) 
      endrows[i]=startrows[i]+(r/p1);
    else
      endrows[i]=startrows[i]+(r/p1)-1;
  }   

  for (int i=0; i<p2; i++)
  {
    if (i<=t%p2) 
      startcols[i]=1+i*((t/p2)+1);
    else
      startcols[i]=1+(t%p2)*((t/p2)+1)+(i-(t%p2))*(t/p2);
    if (i<t%p2) 
      endcols[i]=startcols[i]+(t/p2);
    else
      endcols[i]=startcols[i]+(t/p2)-1;
  }   

  mystartrow=startrows[myprocrow];
  myendrow=endrows[myprocrow];
  mystartcol=startcols[myproccol];
  myendcol=endcols[myproccol];

  MyA=imatrix(mystartrow,myendrow,1,s);
  MyB=imatrix(1,s,mystartcol,myendcol);
  MyR=imatrix(mystartrow,myendrow,mystartcol,myendcol);
  MyA=0.0;
  MyB=0.0;
  
  int actproc;
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
         actproc=startproc+i*p1+j;
         if ((actproc)!=root)
         {
           commerrc=MPI_Send(A,startrows[j],endrows[j],1,s,actproc, 0, MPI_COMM_WORLD);
           commerrc=MPI_Send(B,1,s,startcols[i],endcols[i],actproc, 0, MPI_COMM_WORLD);
         }
         else 
         { 
           for (int k=mystartrow; k<=myendrow; k++)
             MyA[Row(k)]=A[Row(k)];
           for (int k=mystartcol; k<=myendcol; k++)
             MyB[Col(k)]=B[Col(k)];
         }
      }
  }
  else
  {
      commerrc=MPI_Recv(MyA, mystartrow, myendrow, 1, s, root, 
                     0, MPI_COMM_WORLD, &status); 

      commerrc=MPI_Recv(MyB, 1, s, mystartcol, myendcol, root, 
                     0, MPI_COMM_WORLD, &status); 
  }

  //----------------------------------------------------------------------------
  // Start multiplication
  //----------------------------------------------------------------------------

  idotprecision iaccu;
  for (int i=mystartrow; i<=myendrow; i++)
    for (int j=mystartcol; j<=myendcol; j++)
    {   
      iaccu=0.0;
      accumulate(iaccu,MyA[Row(i)],MyB[Col(j)]);
      if (mypid!=root)
        rnd(iaccu,MyR[i][j]);
      else 
        rnd(iaccu,R[i][j]);
    }
  
  //----------------------------------------------------------------------------
  // Gather result
  //----------------------------------------------------------------------------
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
        actproc=startproc+i*p1+j;
        if ((actproc)!=root)
        {

          commerrc=MPI_Recv(R,startrows[j],endrows[j],startcols[i],endcols[i],
                          actproc, 0, MPI_COMM_WORLD, &status);
        }
        else
        {
          // nothing
        }
      }
  }
  else
  {
    commerrc=MPI_Send(MyR, mystartrow, myendrow, mystartcol, myendcol, root, 
                         0, MPI_COMM_WORLD); 
  }
   
  errc = !ok; 

  if (ok)
    ausg << "Matrix multiplication ok. " << endl;
  else
    ausg << "Matrix multiplication failed. " << endl;

}

// ----------------------------------------------------------------------------

/**
 * Matrix multiplication rmatrix*rmatrix A:r x s, B:s x t
 * 
 * @param[in]   A         Input matrix
 * @param[in]   B         Input matrix
 * @param[out]  R         Result
 * @param[in]   r         Matrix dimension 1
 * @param[in]   s         Matrix dimension 2
 * @param[in]   t         Matrix dimension 3
 * @param[in]   root      Master process
 * @param[in]   startproc ID of first involved proc (to date: expected to be =root)
 * @param[in]   endproc   ID of last involved proc
 * @param[out]  errc      Error code
 * @param[out]  commerrc  Error code
 * @param[in,out] ausg    Output file stream
 */
void MatMul(rmatrix& A, rmatrix& B, rmatrix& R, int r, int s, int t,
            int root, int startproc, int endproc,  
            int& errc, int& commerrc, ofstream& ausg) 
{
  //----------------------------------------------------------------------------
  // MPI preparation
  //----------------------------------------------------------------------------

  MPI_Status status;
  int procs=endproc-startproc;
  int mypid; 

  commerrc=MPI_Comm_rank(MPI_COMM_WORLD, &mypid);

  
  ausg << "Start Matrix multiplication..." << endl;

  const real eps = 1e-15;
  bool ok=true;

  SetLb(A,1,1);
  SetLb(A,2,1);
  SetLb(B,1,1);
  SetLb(B,2,1);

  if (mypid==root)
    Resize(R,1,r,1,t); // Result matrix        
  
  rmatrix MyA, MyB, MyR;

  if (mypid==root)
    ok = ((s==(Ub(A,2)-Lb(A,2)+1))&&(s==(Ub(B,1)-Lb(B,1)+1)))? true : false; 
  
  if ((root<startproc)||(root>endproc))
  { 
    commerrc=MPI_ERR_DIMS; 
    return;
  }

  //----------------------------------------------------------------------------
  // Column distribution:
  //----------------------------------------------------------------------------

  // Determine number of procs for rectangular block subdivision of matrices.

  int p1=(int)floor(sqrt((double)procs));
  int p2=p1;
  while (p1*(p2+1)<=procs) 
    p2++;
  endproc=startproc+p1*p2-1;
  
  if ((mypid<startproc)||(mypid>endproc))
    return;
  
  int myproccol=(mypid-startproc)/p1;
  int myprocrow=((mypid-startproc)%p1);

  intvector startrows(0,p1-1);
  intvector endrows(0,p1-1);
  intvector startcols(0,p2-1);
  intvector endcols(0,p2-1);
  
  int mystartrow, myendrow, mystartcol, myendcol;

  // Determine startrows, endrows, startcolumns, endcolumns

  for (int i=0; i<p1; i++)
  {
    if (i<=r%p1) 
      startrows[i]=1+i*((r/p1)+1);
    else
      startrows[i]=1+(r%p1)*((r/p1)+1)+(i-(r%p1))*(r/p1);
    if (i<r%p1) 
      endrows[i]=startrows[i]+(r/p1);
    else
      endrows[i]=startrows[i]+(r/p1)-1;
  }   

  for (int i=0; i<p2; i++)
  {
    if (i<=t%p2) 
      startcols[i]=1+i*((t/p2)+1);
    else
      startcols[i]=1+(t%p2)*((t/p2)+1)+(i-(t%p2))*(t/p2);
    if (i<t%p2) 
      endcols[i]=startcols[i]+(t/p2);
    else
      endcols[i]=startcols[i]+(t/p2)-1;
  }   

  mystartrow=startrows[myprocrow];
  myendrow=endrows[myprocrow];
  mystartcol=startcols[myproccol];
  myendcol=endcols[myproccol];

  MyA=rmatrix(mystartrow,myendrow,1,s);
  MyB=rmatrix(1,s,mystartcol,myendcol);
  MyR=rmatrix(mystartrow,myendrow,mystartcol,myendcol);
  MyA=0.0;
  MyB=0.0;
  
  int actproc;
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
         actproc=startproc+i*p1+j;
         if ((actproc)!=root)
         {
           commerrc=MPI_Send(A,startrows[j],endrows[j],1,s,actproc, 0, MPI_COMM_WORLD);
           commerrc=MPI_Send(B,1,s,startcols[i],endcols[i],actproc, 0, MPI_COMM_WORLD);
         }
         else 
         { 
           for (int k=mystartrow; k<=myendrow; k++)
             MyA[Row(k)]=A[Row(k)];
           for (int k=mystartcol; k<=myendcol; k++)
             MyB[Col(k)]=B[Col(k)];
         }
      }
  }
  else
  {

      commerrc=MPI_Recv(MyA, mystartrow, myendrow, 1, s, root, 
                     0, MPI_COMM_WORLD, &status); 
      commerrc=MPI_Recv(MyB, 1, s, mystartcol, myendcol, root, 
                     0, MPI_COMM_WORLD, &status); 
  }

  //----------------------------------------------------------------------------
  // Start multiplication
  //----------------------------------------------------------------------------

  dotprecision accu;
  for (int i=mystartrow; i<=myendrow; i++)
    for (int j=mystartcol; j<=myendcol; j++)
    {   
      accu=0.0;
      accumulate(accu,MyA[Row(i)],MyB[Col(j)]);
      if (mypid!=root)
        rnd(accu,MyR[i][j]);
      else 
        rnd(accu,R[i][j]);
    }
  
  //----------------------------------------------------------------------------
  // Gather result
  //----------------------------------------------------------------------------
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
        actproc=startproc+i*p1+j;
        if ((actproc)!=root)
        {
          commerrc=MPI_Recv(R,startrows[j],endrows[j],startcols[i],endcols[i],
                          actproc, 0, MPI_COMM_WORLD, &status);
        }
        else
        {
          // nothing
        }
      }
  }
  else
  {
      commerrc=MPI_Send(MyR, mystartrow, myendrow, mystartcol, myendcol, root, 
                         0, MPI_COMM_WORLD); 
  }
   
  errc = !ok; 

  if (ok)
    ausg << "Matrix multiplication ok. " << endl;
  else
    ausg << "Matrix multiplication failed. " << endl;

}


// ----------------------------------------------------------------------------

/**
 * I-A*B for rmatrix*imatrix; A:r x s, B:s x t
 * 
 * @param[in]   A         Input matrix
 * @param[in]   B         Input matrix
 * @param[out]  R         Result
 * @param[in]   r         Matrix dimension 1
 * @param[in]   s         Matrix dimension 2
 * @param[in]   t         Matrix dimension 3
 * @param[in]   root      Master process
 * @param[in]   startproc ID of first involved proc (to date: expected to be =root)
 * @param[in]   endproc   ID of last involved proc
 * @param[out]  errc      Error code
 * @param[out]  commerrc  Error code
 * @param[in,out] ausg    Output file stream
 */
void IminusAB(rmatrix& A, imatrix& B, imatrix& R, int r, int s, int t,
            int root, int startproc, int endproc,  
            int& errc, int& commerrc, ofstream& ausg) 
{
  //----------------------------------------------------------------------------
  // MPI preparation
  //----------------------------------------------------------------------------

  MPI_Status status;
  int procs=endproc-startproc;
  int mypid; 

  commerrc=MPI_Comm_rank(MPI_COMM_WORLD, &mypid);


  ausg << "Start I minus A*B..." << endl;

  const real eps = 1e-15;
  bool ok=true;

  SetLb(A,1,1);
  SetLb(A,2,1);
  SetLb(B,1,1);
  SetLb(B,2,1);

  if (mypid==root)
    Resize(R,1,r,1,t); // Result matrix        
  
  imatrix MyB, MyR;
  rmatrix MyA;

  if (mypid==root)
    ok = ((s==(Ub(A,2)-Lb(A,2)+1))&&(s==(Ub(B,1)-Lb(B,1)+1)))? true : false; 
  
  if ((root<startproc)||(root>endproc))
  { 
    commerrc=MPI_ERR_DIMS; 
    return;
  }

  //----------------------------------------------------------------------------
  // Column distribution:
  //----------------------------------------------------------------------------
  
  // Determine number of procs for rectangular block subdivision of matrices.

  int p1=(int)floor(sqrt((double)procs));
  int p2=p1;
  while (p1*(p2+1)<=procs) 
    p2++;
  endproc=startproc+p1*p2-1;
  
  if ((mypid<startproc)||(mypid>endproc))
    return;
  
  int myproccol=(mypid-startproc)/p1;
  int myprocrow=((mypid-startproc)%p1);
  
  intvector startrows(0,p1-1);
  intvector endrows(0,p1-1);
  intvector startcols(0,p2-1);
  intvector endcols(0,p2-1);
  
  int mystartrow, myendrow, mystartcol, myendcol;

  // Determine startrows, endrows, startcolumns, endcolumns

  for (int i=0; i<p1; i++)
  {
    if (i<=r%p1) 
      startrows[i]=1+i*((r/p1)+1);
    else
      startrows[i]=1+(r%p1)*((r/p1)+1)+(i-(r%p1))*(r/p1);
    if (i<r%p1) 
      endrows[i]=startrows[i]+(r/p1);
    else
      endrows[i]=startrows[i]+(r/p1)-1;
  }   

  for (int i=0; i<p2; i++)
  {
    if (i<=t%p2) 
      startcols[i]=1+i*((t/p2)+1);
    else
      startcols[i]=1+(t%p2)*((t/p2)+1)+(i-(t%p2))*(t/p2);
    if (i<t%p2) 
      endcols[i]=startcols[i]+(t/p2);
    else
      endcols[i]=startcols[i]+(t/p2)-1;
  }   

  mystartrow=startrows[myprocrow];
  myendrow=endrows[myprocrow];
  mystartcol=startcols[myproccol];
  myendcol=endcols[myproccol];

  MyA=rmatrix(mystartrow,myendrow,1,s);
  MyB=imatrix(1,s,mystartcol,myendcol);
  MyR=imatrix(mystartrow,myendrow,mystartcol,myendcol);
  MyA=0.0;
  MyB=0.0;
  
  int actproc;
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
         actproc=startproc+i*p1+j;
         if ((actproc)!=root)
         {
           commerrc=MPI_Send(A,startrows[j],endrows[j],1,s,actproc, 0, MPI_COMM_WORLD);
           commerrc=MPI_Send(B,1,s,startcols[i],endcols[i],actproc, 0, MPI_COMM_WORLD);
         }
         else 
         { 
           for (int k=mystartrow; k<=myendrow; k++)
             MyA[Row(k)]=A[Row(k)];
           for (int k=mystartcol; k<=myendcol; k++)
             MyB[Col(k)]=B[Col(k)];
         }
      }
  }
  else
  {
      commerrc=MPI_Recv(MyA, mystartrow, myendrow, 1, s, root, 
                     0, MPI_COMM_WORLD, &status); 
      commerrc=MPI_Recv(MyB, 1, s, mystartcol, myendcol, root, 
                     0, MPI_COMM_WORLD, &status); 
  }

  //----------------------------------------------------------------------------
  // Start multiplication
  //----------------------------------------------------------------------------

  idotprecision iaccu;
  for (int i=mystartrow; i<=myendrow; i++)
    for (int j=mystartcol; j<=myendcol; j++)
    {   
      iaccu=0.0;
      accumulate(iaccu,MyA[Row(i)],MyB[Col(j)]);
      if (mypid!=root)
        // Store -AB[i][j] !!!
        rnd(-iaccu,MyR[i][j]);
      else 
        // Store -AB[i][j] !!!
        rnd(-iaccu,R[i][j]);
    }
  
  //----------------------------------------------------------------------------
  // Gather result
  //----------------------------------------------------------------------------
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
        actproc=startproc+i*p1+j;
        if ((actproc)!=root)
        {
          commerrc=MPI_Recv(R,startrows[j],endrows[j],startcols[i],endcols[i],
                          actproc, 0, MPI_COMM_WORLD, &status);
        }
        else
        {
          // nothing
        }
      }

    //--------------------------
    // Add Identity Matrix:
    //--------------------------

    for (int i=Lb(R,1); i<=Ub(R,1); i++)
      R[i][i]+=interval(1.0);
  }
  else
  {
      commerrc=MPI_Send(MyR, mystartrow, myendrow, mystartcol, myendcol, root, 
                         0, MPI_COMM_WORLD); 
  }
   
  errc = !ok; 

  if (ok)
    ausg << "I minus A*B ok. " << endl;
  else
    ausg << "I minus A*B failed. " << endl;

}

// ----------------------------------------------------------------------------

/**
 * I-A1*B-A2*B for A1,A2 rmatrix, B imatrix; A1/2:r x s, B:s x t
 * 
 * @param[in]   A1        Input matrix
 * @param[in]   A2        Input matrix
 * @param[in]   B         Input matrix
 * @param[out]  R         Result
 * @param[in]   r         Matrix dimension 1
 * @param[in]   s         Matrix dimension 2
 * @param[in]   t         Matrix dimension 3
 * @param[in]   root      Master process
 * @param[in]   startproc ID of first involved proc (to date: expected to be =root)
 * @param[in]   endproc   ID of last involved proc
 * @param[out]  errc      Error code
 * @param[out]  commerrc  Error code
 * @param[in,out] ausg    Output file stream
 */
void IminusA1A2B(rmatrix& A1, rmatrix& A2, imatrix& B, imatrix& R, 
            int r, int s, int t,
            int root, int startproc, int endproc,  
            int& errc, int& commerrc, ofstream& ausg) 
{
  //----------------------------------------------------------------------------
  // MPI preparation
  //----------------------------------------------------------------------------

  MPI_Status status;
  int procs=endproc-startproc;
  int mypid; 

  commerrc=MPI_Comm_rank(MPI_COMM_WORLD, &mypid);


  ausg << "Start I-A1*B-A2*B..." << endl;

  const real eps = 1e-15;
  bool ok=true;

  SetLb(A1,1,1);
  SetLb(A1,2,1);
  SetLb(A2,1,1);
  SetLb(A2,2,1);
  SetLb(B,1,1);
  SetLb(B,2,1);

  if (mypid==root)
    Resize(R,1,r,1,t); // Result matrix        
  
  imatrix MyB, MyR;
  rmatrix MyA1, MyA2;

  if (mypid==root)
    ok = (   (s==(Ub(A1,2)-Lb(A1,2)+1))&&(s==(Ub(B,1)-Lb(B,1)+1))
          && (s==(Ub(A2,2)-Lb(A2,2)+1))
         )? true : false; 
  if ((root<startproc)||(root>endproc))
  { 
    commerrc=MPI_ERR_DIMS; 
    return;
  }

  //----------------------------------------------------------------------------
  // Column distribution:
  //----------------------------------------------------------------------------

  // Determine number of procs for rectangular block subdivision of matrices.

  int p1=(int)floor(sqrt((double)procs));
  int p2=p1;
  while (p1*(p2+1)<=procs) 
    p2++;
  endproc=startproc+p1*p2-1;
  
  if ((mypid<startproc)||(mypid>endproc))
    return;
  
  int myproccol=(mypid-startproc)/p1;
  int myprocrow=((mypid-startproc)%p1);
  
  intvector startrows(0,p1-1);
  intvector endrows(0,p1-1);
  intvector startcols(0,p2-1);
  intvector endcols(0,p2-1);
  
  int mystartrow, myendrow, mystartcol, myendcol;

  // Determine startrows, endrows, startcolumns, endcolumns

  for (int i=0; i<p1; i++)
  {
    if (i<=r%p1) 
      startrows[i]=1+i*((r/p1)+1);
    else
      startrows[i]=1+(r%p1)*((r/p1)+1)+(i-(r%p1))*(r/p1);
    if (i<r%p1) 
      endrows[i]=startrows[i]+(r/p1);
    else
      endrows[i]=startrows[i]+(r/p1)-1;
  }   

  for (int i=0; i<p2; i++)
  {
    if (i<=t%p2) 
      startcols[i]=1+i*((t/p2)+1);
    else
      startcols[i]=1+(t%p2)*((t/p2)+1)+(i-(t%p2))*(t/p2);
    if (i<t%p2) 
      endcols[i]=startcols[i]+(t/p2);
    else
      endcols[i]=startcols[i]+(t/p2)-1;
  }   

  mystartrow=startrows[myprocrow];
  myendrow=endrows[myprocrow];
  mystartcol=startcols[myproccol];
  myendcol=endcols[myproccol];

  MyA1=rmatrix(mystartrow,myendrow,1,s);
  MyA2=rmatrix(mystartrow,myendrow,1,s);
  MyB=imatrix(1,s,mystartcol,myendcol);
  MyR=imatrix(mystartrow,myendrow,mystartcol,myendcol);
  MyA1=0.0;
  MyA2=0.0;
  MyB=0.0;
  
  int actproc;
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
         actproc=startproc+i*p1+j;
         if ((actproc)!=root)
         {
           commerrc=MPI_Send(A1,startrows[j],endrows[j],1,s,actproc, 0, MPI_COMM_WORLD);
           commerrc=MPI_Send(A2,startrows[j],endrows[j],1,s,actproc, 0, MPI_COMM_WORLD);
           commerrc=MPI_Send(B,1,s,startcols[i],endcols[i],actproc, 0, MPI_COMM_WORLD);
         }
         else 
         { 
           for (int k=mystartrow; k<=myendrow; k++)
             MyA1[Row(k)]=A1[Row(k)];
           for (int k=mystartrow; k<=myendrow; k++)
             MyA2[Row(k)]=A2[Row(k)];
           for (int k=mystartcol; k<=myendcol; k++)
             MyB[Col(k)]=B[Col(k)];
         }
      }
  }
  else
  {
      commerrc=MPI_Recv(MyA1, mystartrow, myendrow, 1, s, root, 
                     0, MPI_COMM_WORLD, &status); 
      commerrc=MPI_Recv(MyA2, mystartrow, myendrow, 1, s, root, 
                     0, MPI_COMM_WORLD, &status); 
      commerrc=MPI_Recv(MyB, 1, s, mystartcol, myendcol, root, 
                     0, MPI_COMM_WORLD, &status); 
  }

  //----------------------------------------------------------------------------
  // Start multiplication
  //----------------------------------------------------------------------------

  idotprecision iaccu;
  for (int i=mystartrow; i<=myendrow; i++)
    for (int j=mystartcol; j<=myendcol; j++)
    {   
      iaccu=0.0;
      accumulate(iaccu,MyA1[Row(i)],MyB[Col(j)]);
      accumulate(iaccu,MyA2[Row(i)],MyB[Col(j)]);
      if (mypid!=root)
        rnd(-iaccu,MyR[i][j]);
      else 
        rnd(-iaccu,R[i][j]);
    }
  
  //----------------------------------------------------------------------------
  // Gather result
  //----------------------------------------------------------------------------
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
        actproc=startproc+i*p1+j;
        if ((actproc)!=root)
        {
          commerrc=MPI_Recv(R,startrows[j],endrows[j],startcols[i],endcols[i],
                          actproc, 0, MPI_COMM_WORLD, &status);
        }
        else
        {
          // nothing
        }
      }

    //--------------------------
    // Add Identity Matrix:
    //--------------------------

    for (int i=Lb(R,1); i<=Ub(R,1); i++)
      R[i][i]+=interval(1.0);
  }
  else
  {
      commerrc=MPI_Send(MyR, mystartrow, myendrow, mystartcol, myendcol, root, 
                         0, MPI_COMM_WORLD); 
  }
   
  errc = !ok; 

  if (ok)
    ausg << "I-A1*B-A2*B ok. " << endl;
  else
    ausg << "I-A1*B-A2*B failed. " << endl;

}



//============================================================================
// #*(A*B-#*(A*B)) for rmatrices
//============================================================================

// ----------------------------------------------------------------------------

/**
 * C1+C2:=#*(A*B-#*(A*B)) for rmatrices; A:r x s, B:s x t
 * 
 * @param[in]   A         Input matrix
 * @param[in]   B         Input matrix
 * @param[out]  C1        Result
 * @param[out]  C2        Result
 * @param[in]   r         Matrix dimension 1
 * @param[in]   s         Matrix dimension 2
 * @param[in]   t         Matrix dimension 3
 * @param[in]   root      Master process
 * @param[in]   startproc ID of first involved proc (to date: expected to be =root)
 * @param[in]   endproc   ID of last involved proc
 * @param[out]  errc      Error code
 * @param[out]  commerrc  Error code
 * @param[in,out] ausg    Output file stream
 */
void ABminusRndAB(rmatrix& A, rmatrix& B, rmatrix& C1, rmatrix& C2, 
            int r, int s, int t,
            int root, int startproc, int endproc,  
            int& errc, int& commerrc, ofstream& ausg) 
{
  //----------------------------------------------------------------------------
  // MPI preparation
  //----------------------------------------------------------------------------

  MPI_Status status;
  int procs=endproc-startproc;
  int mypid;

  commerrc=MPI_Comm_rank(MPI_COMM_WORLD, &mypid);

  
  ausg << "Start #*(A*B-#*(A*B))..." << endl;

  const real eps = 1e-15;
  bool ok=true;

  SetLb(A,1,1);
  SetLb(A,2,1);
  SetLb(B,1,1);
  SetLb(B,2,1);

  if (mypid==root)
  {
    Resize(C1,1,r,1,t); // Result matrix        
    Resize(C2,1,r,1,t); // Result matrix        
  }
  
  rmatrix MyB, MyC1, MyC2;
  rmatrix MyA;

  if (mypid==root)
  {
    ok = ((r==(Ub(A,1)-Lb(A,1)+1))&&(t==(Ub(B,2)-Lb(B,2)+1))
          &&(s==(Ub(A,2)-Lb(A,2)+1))&&(s==(Ub(B,1)-Lb(B,1)+1))
         )? true : false; 
  }
  
  if ((root<startproc)||(root>endproc))
  { 
    commerrc=MPI_ERR_DIMS; 
    return;
  }

  //----------------------------------------------------------------------------
  // Column distribution:
  //----------------------------------------------------------------------------

  // Determine number of procs for rectangular block subdivision of matrices.

  int p1=(int)floor(sqrt((double)procs));
  int p2=p1;
  while (p1*(p2+1)<=procs) 
    p2++;
  endproc=startproc+p1*p2-1;
  
  if ((mypid<startproc)||(mypid>endproc))
    return;
  
  int myproccol=(mypid-startproc)/p1;
  int myprocrow=((mypid-startproc)%p1);

  intvector startrows(0,p1-1);
  intvector endrows(0,p1-1);
  intvector startcols(0,p2-1);
  intvector endcols(0,p2-1);
  
  int mystartrow, myendrow, mystartcol, myendcol;

  // Determine startrows, endrows, startcolumns, endcolumns

  for (int i=0; i<p1; i++)
  {
    if (i<=r%p1) 
      startrows[i]=1+i*((r/p1)+1);
    else
      startrows[i]=1+(r%p1)*((r/p1)+1)+(i-(r%p1))*(r/p1);
    if (i<r%p1) 
      endrows[i]=startrows[i]+(r/p1);
    else
      endrows[i]=startrows[i]+(r/p1)-1;
  }   

  for (int i=0; i<p2; i++)
  {
    if (i<=t%p2) 
      startcols[i]=1+i*((t/p2)+1);
    else
      startcols[i]=1+(t%p2)*((t/p2)+1)+(i-(t%p2))*(t/p2);
    if (i<t%p2) 
      endcols[i]=startcols[i]+(t/p2);
    else
      endcols[i]=startcols[i]+(t/p2)-1;
  }   

  mystartrow=startrows[myprocrow];
  myendrow=endrows[myprocrow];
  mystartcol=startcols[myproccol];
  myendcol=endcols[myproccol];

  MyA=rmatrix(mystartrow,myendrow,1,s);
  MyB=rmatrix(1,s,mystartcol,myendcol);
  MyC1=rmatrix(mystartrow,myendrow,mystartcol,myendcol);
  MyC2=rmatrix(mystartrow,myendrow,mystartcol,myendcol);
  MyA=0.0;
  MyB=0.0;
  MyC1=0.0;
  MyC2=0.0;
  
  int actproc;
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
         actproc=startproc+i*p1+j;
         if ((actproc)!=root)
         {
           commerrc=MPI_Send(A,startrows[j],endrows[j],1,s,actproc, 0, MPI_COMM_WORLD);
           commerrc=MPI_Send(B,1,s,startcols[i],endcols[i],actproc, 0, MPI_COMM_WORLD);
         }
         else 
         { 
           for (int k=mystartrow; k<=myendrow; k++)
             MyA[Row(k)]=A[Row(k)];
           for (int k=mystartcol; k<=myendcol; k++)
             MyB[Col(k)]=B[Col(k)];
         }
      }
  }
  else
  {
      commerrc=MPI_Recv(MyA, mystartrow, myendrow, 1, s, root, 
                     0, MPI_COMM_WORLD, &status); 
      commerrc=MPI_Recv(MyB, 1, s, mystartcol, myendcol, root, 
                     0, MPI_COMM_WORLD, &status); 

  }

  //----------------------------------------------------------------------------
  // Start multiplication
  //----------------------------------------------------------------------------

  dotprecision accu;
  for (int i=mystartrow; i<=myendrow; i++)
    for (int j=mystartcol; j<=myendcol; j++)
    {   
      // Entry [i,j] of A*B-C:
      
      accu=0.0;
      accumulate(accu,MyA[Row(i)],MyB[Col(j)]);
      
      if (mypid!=root)
      {
        rnd(accu,MyC1[i][j]);
        accu-=MyC1[i][j];
        rnd(accu,MyC2[i][j]);
      }
      else 
      {
        rnd(accu,C1[i][j]);
        accu-=C1[i][j];
        rnd(accu,C2[i][j]);
      } 
    }
  
  //----------------------------------------------------------------------------
  // Gather result
  //----------------------------------------------------------------------------
  
  if (mypid==root)
  {
    for (int i=0; i<p2; i++)
      for (int j=0; j<p1; j++)
      {
        actproc=startproc+i*p1+j;
        if ((actproc)!=root)
        {
          commerrc=MPI_Recv(C1,startrows[j],endrows[j],startcols[i],endcols[i],
                          actproc, 0, MPI_COMM_WORLD, &status);
          commerrc=MPI_Recv(C2,startrows[j],endrows[j],startcols[i],endcols[i],
                          actproc, 0, MPI_COMM_WORLD, &status);
        }
        else
        {
          // nothing
        }
      }
  }
  else
  {
      commerrc=MPI_Send(MyC1, mystartrow, myendrow, mystartcol, myendcol, root, 
                         0, MPI_COMM_WORLD); 
      commerrc=MPI_Send(MyC2, mystartrow, myendrow, mystartcol, myendcol, root, 
                         0, MPI_COMM_WORLD); 
  }
   
  errc = !ok; 

  if (ok)
    ausg << "#*(A*B-#*(A*B)) ok. " << endl;
  else
    ausg << "#*(A*B-#*(A*B)) failed. " << endl;

}



