/*
**  CXSC is a C++ library for eXtended Scientific Computing 
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany
**
**  MPI communication package for C-XSC Taylor arithmetic data types
**
**  Copyright (C) 2004-2007 Markus Grimmer
**
**  markus.grimmer@math.uni-wuppertal.de
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @file taylor_mpicomm.cpp
 * 
 * MPI communication package for C-XSC Taylor arithmetic data types: 
 * Implementation file.
 *
 * Copyright (C) 2004-2007 Markus Grimmer
 *
 * Incorporated Routines:
 *
 * <ul>
 * <li> MPI_Pack, MPI_Unpack </li>
 * </ul>
 *
 * Incorporated Types:
 *
 * <ul>
 * <li> itaylor, dim2taylor, dim2taylor_vector </li>
 * </ul>
 */


#include "cxsc_mpicomm_templ.hpp"
#include "taylor_mpicomm.hpp"

using namespace cxsc;
using namespace std;
using namespace taylor;


// =============================================================================
//
// Taylor data types
//
// =============================================================================

// =============================================================================
// MPI-Send/Recv for itaylor:
// =============================================================================

int MPI_Pack (itaylor& it, void* sendbuffer, int bufflen, int* pos,  MPI_Comm MC)
{
  int err;
  int p;
  int lb, ub;
  
  if (!MPI_CXSC_TYPES_DEFINED) 
     if (err=MPI_Define_CXSC_Types()!=MPI_SUCCESS)
        return err;

  p = it.order();
  
  //  err=MPI_Pack(&r,1,MPI_CXSC_REAL,buff,bufflen,pos,MC);

  
  if (err=MPI_Pack(&p,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
      !=MPI_SUCCESS)
    return err;
  if (err=MPI_Pack(&(it[0]),p+1,MPI_CXSC_INTERVAL,sendbuffer,
                   MPI_CXSC_BUFFERLEN,pos,MC)
      !=MPI_SUCCESS)
    return err;

  return err;
}

int MPI_Unpack (void* recbuffer, int bufflen, int* pos, itaylor& it, MPI_Comm MC)
{
  int err;
  int vlen, lb, ub, p;

  if (!MPI_CXSC_TYPES_DEFINED) 
     if (err=MPI_Define_CXSC_Types()!=MPI_SUCCESS)
        return err;

  if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &p, 1, MPI_INT, MC)
      !=MPI_SUCCESS)  
      return err;

  itaylor tmpit(p);
 
  if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &tmpit[0], p+1, 
      MPI_CXSC_INTERVAL, MC)
      !=MPI_SUCCESS)  
      return err;
             
  it=tmpit;
  return err;
}

// =============================================================================
// MPI-Send/Recv for dim2taylor:
// =============================================================================


int MPI_Pack (dim2taylor& d2t, void* sendbuffer, int bufflen, int* pos,  MPI_Comm MC)
{
  int err;
  int p;
  int lb, ub;
  
  if (!MPI_CXSC_TYPES_DEFINED) 
     if (err=MPI_Define_CXSC_Types()!=MPI_SUCCESS)
        return err;

  p = d2t.get_p();
  if (err=MPI_Pack(&p,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
      !=MPI_SUCCESS)
      return err;
  
  for (int i=0; i<=p; i++)  // p+1 ivectors
    {
      lb=Lb(d2t[i]);
      ub=Ub(d2t[i]);  
      if (err=MPI_Pack(&lb,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
          !=MPI_SUCCESS)
        return err;
      if (err=MPI_Pack(&ub,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
          !=MPI_SUCCESS)
        return err;
      if (err=MPI_Pack(&(d2t[i][lb]),ub-lb+1,MPI_CXSC_INTERVAL,sendbuffer,
          MPI_CXSC_BUFFERLEN,pos,MC)
          !=MPI_SUCCESS)
        return err;
    }

  return err;
}

int MPI_Unpack (void* recbuffer, int bufflen, int* pos, dim2taylor& d2t, MPI_Comm MC)
{
  int err;
  int vlen, lb, ub, p;
  ivector tmpelem;

  if (!MPI_CXSC_TYPES_DEFINED) 
     if (err=MPI_Define_CXSC_Types()!=MPI_SUCCESS)
        return err;

  if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &p, 1, MPI_INT, MC)
      !=MPI_SUCCESS)     
      return err;

  dim2taylor tmpd2t(p);
 
  for (int i=0; i<=p; i++)
    {
      if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &lb, 
          1, MPI_INT, MC) !=MPI_SUCCESS)
          return err;
      if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &ub, 
          1, MPI_INT, MC) !=MPI_SUCCESS)
          return err;
      vlen=ub-lb+1;
      Resize(tmpelem,vlen);
      if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &tmpelem[1] , 
          vlen, MPI_CXSC_INTERVAL, MC) !=MPI_SUCCESS)
          return err;
          
      SetLb(tmpelem,lb);  // ub not necessary (automatically set)
      tmpd2t[i]=tmpelem;
    } 
  d2t=tmpd2t;
  return err;
}


// =============================================================================
// MPI-Send/Recv for dim2taylor_vector:
// =============================================================================

int MPI_Pack (dim2taylor_vector& d2tv, void* sendbuffer, 
              int bufflen, int* pos,  MPI_Comm MC)
{
  int err;
  int pv, p;
  int lb, ub, lbv, ubv;
  
  if (!MPI_CXSC_TYPES_DEFINED) 
     if (err=MPI_Define_CXSC_Types()!=MPI_SUCCESS)
        return err;

  pv = d2tv.get_p_el();
  lbv = Lb(d2tv);
  ubv = Ub(d2tv);

  if (err=MPI_Pack(&lbv,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
      !=MPI_SUCCESS)
        return err;
  if (err=MPI_Pack(&ubv,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
      !=MPI_SUCCESS)
        return err;
  if (err=MPI_Pack(&pv,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC) 
      !=MPI_SUCCESS)
        return err;

  for (int j=lbv; j<=ubv; j++)
    {
      //pack dim2taylor objects
      
      p = d2tv[j].get_p();
      if (err=MPI_Pack(&p,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
        !=MPI_SUCCESS)
          return err;
 
      for (int i=0; i<=p; i++)  // p+1 ivectors
        {
          lb=Lb(d2tv[j][i]);
          ub=Ub(d2tv[j][i]);
          if (err=MPI_Pack(&lb,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
              !=MPI_SUCCESS)
              return err;
          if (err=MPI_Pack(&ub,1,MPI_INT,sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
              !=MPI_SUCCESS)
              return err;
          if (err=MPI_Pack(&(d2tv[j][i][lb]),ub-lb+1,MPI_CXSC_INTERVAL,
              sendbuffer,MPI_CXSC_BUFFERLEN,pos,MC)
              !=MPI_SUCCESS)
              return err;
        }
    }
  return err;
}

int MPI_Unpack (void* recbuffer, int bufflen, int* pos,
              dim2taylor_vector& d2tv, MPI_Comm MC)
{
  int err;
  int pv, p;
  int lb, ub, lbv, ubv, vlen;
  ivector tmpelem;

  if (!MPI_CXSC_TYPES_DEFINED) 
     if (err=MPI_Define_CXSC_Types()!=MPI_SUCCESS)
        return err;

  if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &lbv, 1, MPI_INT, MC)
      !=MPI_SUCCESS)
      return err;
  if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &ubv, 1, MPI_INT, MC)
      !=MPI_SUCCESS)
      return err;
  if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &pv, 1, MPI_INT, MC)  
      !=MPI_SUCCESS)
      return err;

  dim2taylor_vector tmpd2tv(pv,lbv,ubv);
  
  for (int j=lbv; j<=ubv; j++)
    {
      if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &p, 1, MPI_INT, MC)
         !=MPI_SUCCESS)
         return err;

      dim2taylor tmpd2t(p);
 
      for (int i=0; i<=p; i++)
        {
          if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &lb,
              1, MPI_INT, MC) !=MPI_SUCCESS)
              return err;

          if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &ub,
              1, MPI_INT, MC) !=MPI_SUCCESS)
              return err;

          vlen=ub-lb+1;
          Resize(tmpelem,vlen);
          if (err=MPI_Unpack(recbuffer, MPI_CXSC_BUFFERLEN, pos, &tmpelem[1] , 
              vlen, MPI_CXSC_INTERVAL, MC) !=MPI_SUCCESS)
              return err;

          SetLb(tmpelem,lb);  // ub not necessary (automatically set)
          tmpd2t[i]=tmpelem;
        }
      tmpd2tv[j]=tmpd2t;
    }
  d2tv=tmpd2tv;

  return err;
}


