/*
**  CXSC is a C++ library for eXtended Scientific Computing 
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany
**
**  MPI communication package for STL vectors
**  of C-XSC data types and C-XSC Taylor arithmetic types
**
**  Copyright (C) 2004-2007 Markus Grimmer
**
**  markus.grimmer@math.uni-wuppertal.de
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @file vector_mpicomm.cpp
 * 
 * MPI communication package for STL vectors
 * of C-XSC data types and C-XSC Taylor arithmetic types:
 * Implementation file.
 *
 * Copyright (C) 2004-2007 Markus Grimmer
 *
 * Incorporated Routines:
 *
 * <ul>
 * <li> (see module cxsc_mpicomm_templ) </li>
 * </ul>
 *
 * Incorporated Types:
 *
 * <ul>
 * <li> vector </li>
 * </ul>
 */

// Remark: templates have to be explicitly instantiated!
// Remark: This strategy should work for all element data types that can be sent
//         as MPI_Send(Pointer,int,int,MPI_Comm)!
//         MPI_Send originally has the interface
//         MPI_Send (Pointer,int,MPI_Datatype,int,int,MPI_Comm).

#include<iostream> //Test purposes

#include "cxsc_mpicomm_templ.hpp"
#include "vector_mpicomm.hpp"

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// Addition for taylor types: 
// (also see explicit instantiation at end of file)
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#include "taylor_mpicomm.hpp"

using namespace taylor;


using namespace std;
using namespace cxsc;

/**
 * Communication buffer.
 */
extern char commbuffer[MPI_CXSC_BUFFERLEN]; 

// =============================================================================
// MPI-Pack/Unpack for vector<T>:
// =============================================================================


template<class T>
int MPI_Pack (vector<T>& tv, void* buff, int bufflen, int* pos,  MPI_Comm MC)
{
  int err;
  int len=tv.size();

  if (err=MPI_Pack(&len,1,MPI_INT,buff,bufflen,pos,MC)!=MPI_SUCCESS)
    return err;

  for (int i=0; i<len; i++)
    {
      if (err=MPI_Pack(tv[i],buff,bufflen,pos,MC)!= MPI_SUCCESS)
      return err;
    }

  return err;
}

template<class T>
int MPI_Unpack (void* buff, int bufflen, int* pos, vector<T>& tv, MPI_Comm MC)
{
  int err;
  int len;
  
  if (err=MPI_Unpack(buff, bufflen, pos, &len, 1, MPI_INT, MC)!=MPI_SUCCESS)
    return err;

  tv.resize(len);

  for (int i=0; i<len; i++)
    {
      if (err=MPI_Unpack(buff,bufflen,pos,tv[i],MC)!= MPI_SUCCESS)
      return err;
    }
  return err;
}

// =============================================================================
// Point-to-point Communication
// =============================================================================


template<class T>
int MPI_Send(vector<T>& tv, int i1, int i2, MPI_Comm MC)
{
  // Pack the vector, then send it.

  int pos=0;
  int err;

  // Call overloaded MPI_Pack:

  if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
    return err;
  err=MPI_Send(commbuffer, pos, MPI_PACKED, i1, i2, MC);
  return err;
}

template<class T>
int MPI_Bsend(vector<T>& tv, int i1, int i2, MPI_Comm MC)
{
  // Pack the vector, then send it.

  int pos=0;
  int err;

  // Call overloaded MPI_Pack:

  if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
    return err;
  err=MPI_Bsend(commbuffer, pos, MPI_PACKED, i1, i2, MC);
  return err;
}

template<class T>
int MPI_Ssend(vector<T>& tv, int i1, int i2, MPI_Comm MC)
{
  // Pack the vector, then send it.

  int pos=0;
  int err;

  // Call overloaded MPI_Pack:

  if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
    return err;
  err=MPI_Ssend(commbuffer, pos, MPI_PACKED, i1, i2, MC);
  return err;
}

template<class T>
int MPI_Rsend(vector<T>& tv, int i1, int i2, MPI_Comm MC)
{
  // Pack the vector, then send it.

  int pos=0;
  int err;

  // Call overloaded MPI_Pack:

  if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
    return err;
  err=MPI_Rsend(commbuffer, pos, MPI_PACKED, i1, i2, MC);
  return err;
}

template<class T>
int MPI_Isend(vector<T>& tv, int i1, int i2, MPI_Comm MC, MPI_Request* MR)
{
  // Pack the vector, then send it.

  int pos=0;
  int err;

  // Call overloaded MPI_Pack:

  if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
    return err;
  err=MPI_Isend(commbuffer, pos, MPI_PACKED, i1, i2, MC, MR);
  return err;
}

template<class T>
int MPI_Ibsend(vector<T>& tv, int i1, int i2, MPI_Comm MC, MPI_Request* MR)
{
  // Pack the vector, then send it.

  int pos=0;
  int err;

  // Call overloaded MPI_Pack:

  if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
    return err;
  err=MPI_Ibsend(commbuffer, pos, MPI_PACKED, i1, i2, MC, MR);
  return err;
}

template<class T>
int MPI_Issend(vector<T>& tv, int i1, int i2, MPI_Comm MC, MPI_Request* MR)
{
  // Pack the vector, then send it.

  int pos=0;
  int err;

  // Call overloaded MPI_Pack:

  if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
    return err;
  err=MPI_Issend(commbuffer, pos, MPI_PACKED, i1, i2, MC, MR);
  return err;
}

template<class T>
int MPI_Irsend(vector<T>& tv, int i1, int i2, MPI_Comm MC, MPI_Request* MR)
{
  // Pack the vector, then send it.

  int pos=0;
  int err;

  // Call overloaded MPI_Pack:

  if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
    return err;
  err=MPI_Irsend(commbuffer, pos, MPI_PACKED, i1, i2, MC, MR);
  return err;
}

template<class T>
int MPI_Recv(vector<T>& tv, int i1, int i2, MPI_Comm MC, MPI_Status* MS)
{
  int pos=0;
  int err;

  if (err=MPI_Recv(commbuffer, MPI_CXSC_BUFFERLEN, MPI_PACKED, 
                   i1, i2, MC, MS)!= MPI_SUCCESS)
    return err;

  // Call overloaded MPI_Unpack

  err=MPI_Unpack(commbuffer, MPI_CXSC_BUFFERLEN, &pos, tv, MC);  
  return err;
}


// =============================================================================
// Collective Communication: MPI_Bcast
// =============================================================================


template<class T>
int MPI_Bcast(vector<T>& tv, int root, MPI_Comm MC)
{
  int pos=0;
  int len=0;
  int err;

  int mypid;
  err=MPI_Comm_rank(MC, &mypid);
  
  if (mypid==root)  
  {
    // Pack the vector, then broadcast it.
    // It is assumed that a suitable
    // MPI_Pack is available!

    // Call overloaded MPI_Pack:

    if (err=MPI_Pack(tv, commbuffer, MPI_CXSC_BUFFERLEN, &pos, MC)!=MPI_SUCCESS)
      return err;
    len=pos;
    if (err=MPI_Bcast(&len, 1, MPI_INT, root, MC)!=MPI_SUCCESS)
      return err;
    if (err=MPI_Bcast(commbuffer, len, MPI_PACKED, root, MC)!=MPI_SUCCESS)
      return err;
  }
  else
  {

    // Broadcast-receive the vector, then unpack it.
    // It is assumed that a suitable
    // MPI_Unpack is available!

    if (err=MPI_Bcast(&len, 1, MPI_INT, root, MC)!= MPI_SUCCESS)
      return err;
    if (err=MPI_Bcast(commbuffer, len, MPI_PACKED, root, MC)!= MPI_SUCCESS)
      return err;

    // Call overloaded MPI_Unpack:

    err=MPI_Unpack(commbuffer, MPI_CXSC_BUFFERLEN, &pos, tv, MC);
  }
  
  return err;
}



// =============================================================================
// Specializations
// =============================================================================


template<>
int MPI_Pack<double> (vector<double>& tv, void* buff, int bufflen, int* pos,  MPI_Comm MC)
{
  int err;
  int len=tv.size();

  if (err=MPI_Pack(&len,1,MPI_INT,buff,bufflen,pos,MC)!=MPI_SUCCESS)
    return err;

  for (int i=0; i<len; i++)
    {
      if (err=MPI_Pack(&tv[i],1,MPI_DOUBLE,buff,bufflen,pos,MC)!= MPI_SUCCESS)
      return err;
    }

  return err;
}

template<>
int MPI_Unpack<double> (void* buff, int bufflen, int* pos, vector<double>& tv, MPI_Comm MC)
{
  int err=0;
  int len=0;
  
  if (err=MPI_Unpack(buff, bufflen, pos, &len, 1, MPI_INT, MC)!=MPI_SUCCESS)
    return err;

  tv.resize(len);

  for (int i=0; i<len; i++)
  {
    if (err=MPI_Unpack(buff,bufflen,pos,&tv[i],1,MPI_DOUBLE,MC)!= MPI_SUCCESS)
    return err;
  }
  return err;
}

template<>
int MPI_Unpack<itaylor> (void* buff, int bufflen, int* pos, vector<itaylor>& tv, MPI_Comm MC)
{
  int err;
  int len;
  
  if (err=MPI_Unpack(buff, bufflen, pos, &len, 1, MPI_INT, MC)!=MPI_SUCCESS)
    return err;

  itaylor ittmp;

  for (int i=0; i<len; i++)
    {
      if (err=MPI_Unpack(buff,bufflen,pos,ittmp,MC)!= MPI_SUCCESS)
        return err;

      tv[i]=ittmp;
    }
    
  return err;
}

template<>
int MPI_Unpack<dim2taylor> (void* buff, int bufflen, int* pos, vector<dim2taylor>& tv, MPI_Comm MC)
{
  int err;
  int len;
  
  if (err=MPI_Unpack(buff, bufflen, pos, &len, 1, MPI_INT, MC)!=MPI_SUCCESS)
    return err;

  dim2taylor d2ttmp;

  tv.resize(len);

  for (int i=0; i<len; i++)
    {
      if (err=MPI_Unpack(buff,bufflen,pos,d2ttmp,MC)!= MPI_SUCCESS)
        return err;
      tv[i]=d2ttmp;
    }
    
  return err;
}

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// Explicit instantiation of templates
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

// You'll most probably need explicit instantiation 
// for instances you'd like to use in your program: 

template int MPI_Send<double> (vector<double>&, int, int, MPI_Comm);
template int MPI_Rsend<double> (vector<double>&, int, int, MPI_Comm);
template int MPI_Bsend<double> (vector<double>&, int, int, MPI_Comm);
template int MPI_Ssend<double> (vector<double>&, int, int, MPI_Comm);
template int MPI_Isend<double> (vector<double>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Irsend<double>(vector<double>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Issend<double>(vector<double>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Ibsend<double>(vector<double>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Recv<double>  (vector<double>&, int, int, MPI_Comm, MPI_Status*);
template int MPI_Bcast<double> (vector<double>&, int, MPI_Comm);

template int MPI_Send<rvector> (vector<rvector>&, int, int, MPI_Comm);
template int MPI_Rsend<rvector> (vector<rvector>&, int, int, MPI_Comm);
template int MPI_Bsend<rvector> (vector<rvector>&, int, int, MPI_Comm);
template int MPI_Ssend<rvector> (vector<rvector>&, int, int, MPI_Comm);
template int MPI_Isend<rvector> (vector<rvector>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Irsend<rvector>(vector<rvector>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Issend<rvector>(vector<rvector>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Ibsend<rvector>(vector<rvector>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Recv<rvector>  (vector<rvector>&, int, int, MPI_Comm, MPI_Status*);
template int MPI_Bcast<rvector> (vector<rvector>&, int, MPI_Comm);

template int MPI_Send<ivector> (vector<ivector>&, int, int, MPI_Comm);
template int MPI_Rsend<ivector> (vector<ivector>&, int, int, MPI_Comm);
template int MPI_Bsend<ivector> (vector<ivector>&, int, int, MPI_Comm);
template int MPI_Ssend<ivector> (vector<ivector>&, int, int, MPI_Comm);
template int MPI_Isend<ivector> (vector<ivector>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Irsend<ivector>(vector<ivector>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Issend<ivector>(vector<ivector>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Ibsend<ivector>(vector<ivector>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Recv<ivector>  (vector<ivector>&, int, int, MPI_Comm, MPI_Status*);
template int MPI_Bcast<ivector> (vector<ivector>&, int, MPI_Comm);

template int MPI_Send<vector<ivector> > (vector<vector<ivector> >&, int, int, MPI_Comm);
template int MPI_Rsend<vector<ivector> > (vector<vector<ivector> >&, int, int, MPI_Comm);
template int MPI_Bsend<vector<ivector> > (vector<vector<ivector> >&, int, int, MPI_Comm);
template int MPI_Ssend<vector<ivector> > (vector<vector<ivector> >&, int, int, MPI_Comm);
template int MPI_Isend<vector<ivector> > (vector<vector<ivector> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Irsend<vector<ivector> >(vector<vector<ivector> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Issend<vector<ivector> >(vector<vector<ivector> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Ibsend<vector<ivector> >(vector<vector<ivector> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Recv<vector<ivector> >  (vector<vector<ivector> >&, int, int, MPI_Comm, MPI_Status*);
template int MPI_Bcast<vector<ivector> > (vector<vector<ivector> >&, int, MPI_Comm);

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// Explicit instantiation for taylor types (1- and 2-dimensional
// Taylor Arithmetic) - remove if not applicable
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

template int MPI_Send<itaylor> (vector<itaylor>&, int, int, MPI_Comm);
template int MPI_Rsend<itaylor> (vector<itaylor>&, int, int, MPI_Comm);
template int MPI_Bsend<itaylor> (vector<itaylor>&, int, int, MPI_Comm);
template int MPI_Ssend<itaylor> (vector<itaylor>&, int, int, MPI_Comm);
template int MPI_Isend<itaylor> (vector<itaylor>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Irsend<itaylor>(vector<itaylor>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Issend<itaylor>(vector<itaylor>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Ibsend<itaylor>(vector<itaylor>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Recv<itaylor>  (vector<itaylor>&, int, int, MPI_Comm, MPI_Status*);
template int MPI_Bcast<itaylor> (vector<itaylor>&, int, MPI_Comm);

template int MPI_Send< vector<itaylor> > (vector< vector<itaylor> >&, int, int, MPI_Comm);
template int MPI_Rsend< vector<itaylor> > (vector< vector<itaylor> >&, int, int, MPI_Comm);
template int MPI_Bsend< vector<itaylor> > (vector< vector<itaylor> >&, int, int, MPI_Comm);
template int MPI_Ssend< vector<itaylor> > (vector< vector<itaylor> >&, int, int, MPI_Comm);
template int MPI_Isend< vector<itaylor> > (vector< vector<itaylor> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Irsend< vector<itaylor> >(vector< vector<itaylor> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Issend< vector<itaylor> >(vector< vector<itaylor> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Ibsend< vector<itaylor> >(vector< vector<itaylor> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Recv< vector<itaylor> >  (vector< vector<itaylor> >&, int, int, MPI_Comm, MPI_Status*);
template int MPI_Bcast< vector<itaylor> > (vector< vector<itaylor> >&, int, MPI_Comm);

template int MPI_Send<dim2taylor> (vector<dim2taylor>&, int, int, MPI_Comm);
template int MPI_Rsend<dim2taylor> (vector<dim2taylor>&, int, int, MPI_Comm);
template int MPI_Bsend<dim2taylor> (vector<dim2taylor>&, int, int, MPI_Comm);
template int MPI_Ssend<dim2taylor> (vector<dim2taylor>&, int, int, MPI_Comm);
template int MPI_Isend<dim2taylor> (vector<dim2taylor>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Irsend<dim2taylor>(vector<dim2taylor>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Issend<dim2taylor>(vector<dim2taylor>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Ibsend<dim2taylor>(vector<dim2taylor>&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Recv<dim2taylor>  (vector<dim2taylor>&, int, int, MPI_Comm, MPI_Status*);
template int MPI_Bcast<dim2taylor> (vector<dim2taylor>&, int, MPI_Comm);

template int MPI_Send< vector<dim2taylor> > (vector< vector<dim2taylor> >&, int, int, MPI_Comm);
template int MPI_Rsend< vector<dim2taylor> > (vector< vector<dim2taylor> >&, int, int, MPI_Comm);
template int MPI_Bsend< vector<dim2taylor> > (vector< vector<dim2taylor> >&, int, int, MPI_Comm);
template int MPI_Ssend< vector<dim2taylor> > (vector< vector<dim2taylor> >&, int, int, MPI_Comm);
template int MPI_Isend< vector<dim2taylor> > (vector< vector<dim2taylor> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Irsend< vector<dim2taylor> >(vector< vector<dim2taylor> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Issend< vector<dim2taylor> >(vector< vector<dim2taylor> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Ibsend< vector<dim2taylor> >(vector< vector<dim2taylor> >&, int, int, MPI_Comm, MPI_Request*);
template int MPI_Recv< vector<dim2taylor> >  (vector< vector<dim2taylor> >&, int, int, MPI_Comm, MPI_Status*);
template int MPI_Bcast< vector<dim2taylor> > (vector< vector<dim2taylor> >&, int, MPI_Comm);
